#ifndef DUMMY_PARSER_HPP
#define DUMMY_PARSER_HPP

#include "Lexer.hpp"
#include "Parser.hpp"
#include <assert.h>
#include <map>
#include <sstream>
#include <string>
#include <vector>

enum SymbolType {
    ST_VARIABLE,
    ST_PARAMETER,
    ST_FUNCTION,
    ST_UNDEFINED,
    ST_METADATA
};

enum DataType {
    DT_INTEGER,
    DT_DOUBLE,
    DT_ARBITRARY
};

struct SymbolData {
    SymbolType symbol_type;
    std::vector<DataType> data_type;
};

class SymbolTable {
private:
    std::vector<std::map<std::string, SymbolData> > maps;

public:
    void push()
    {
        maps.emplace_back(std::map<std::string, SymbolData>{});
    }
    void pop()
    {
        assert(!maps.empty());
        maps.pop_back();
    }
    SymbolData& find(std::string name)
    {
        assert(!maps.empty());
        for (std::vector<std::map<std::string, SymbolData> >::reverse_iterator iter = maps.rbegin();
             iter != maps.rend(); ++iter) {
            auto search = iter->find(name);
            if (search != iter->end()) {
                return search->second;
            }
        }
        auto insert_result = maps.back().insert(std::pair<std::string, SymbolData>(
            name, { ST_UNDEFINED, { DT_ARBITRARY } }));
        assert(insert_result.second);
        return insert_result.first->second;
    }
    bool insert(std::string name, SymbolData data)
    {
        assert(!maps.empty());
        return maps.back().insert(std::pair<std::string, SymbolData>(name, data)).second;
    }
};

// This parser writes XML syntax tree produced from the input to Parser::out.
class XMLParser : public Parser {
private:
    // Used to dynamically switch the user-specified Parser::out output stream
    // for string streams to capture the parser output.
    std::vector<std::ostringstream> out_streams{};
    void start_buffering()
    {
        out_streams.emplace_back(std::ostringstream{});
    }
    std::ostringstream stop_buffering()
    {
        std::ostringstream out_stream = std::move(out_streams.back());
        assert(!out_streams.empty());
        out_streams.pop_back();
        return out_stream;
    }
    std::ostream& get_current_out()
    {
        if (out_streams.empty())
            return out;
        else
            return out_streams.back();
    }

    SymbolTable symbols;

    DataType arithmetic_binary_operator_coerce(
        DataType left_operand_type, DataType right_operator_type);
    DataType return_data_type_coerce(
        const Token& where, DataType expected_type, DataType actual_type);

    void grammar_init();
    void grammar_block();
    void grammar_variable_declaration_list(bool is_nested);
    void grammar_variable_declaration();
    DataType grammar_type();
    void grammar_callable_declaration_list(bool is_nested);
    void grammar_callable_declaration();
    void grammar_parameter_list(std::vector<DataType>& types);
    void grammar_parameter_list_tail(std::vector<DataType>& types);
    void grammar_parameter(std::vector<DataType>& types);
    void grammar_statement_list(bool is_nested);
    void grammar_statement();
    void grammar_statement_if_tail();
    void grammar_coerced_arithmetic_expression(const Token& where, DataType expected_type);
    DataType grammar_arithmetic_expression();
    DataType grammar_arithmetic_expression_tail(DataType left_operand_type);
    void grammar_arithmetic_expression_list(std::vector<DataType>& expected_types);
    void grammar_arithmetic_expression_list_tail(std::vector<DataType>& expected_types);
    DataType grammar_arithmetic_factor();
    DataType grammar_arithmetic_factor_tail(DataType left_operand_type);
    DataType grammar_arithmetic_operand();
    void grammar_bool_expression();
    void grammar_bool_expression_tail();
    void grammar_bool_factor();
    void grammar_bool_factor_tail();
    void grammar_bool_operand();
    void grammar_relation_expression();
    void grammar_relation_expression_tail();

public:
    XMLParser(Lexer& lexer, std::ostream& out, std::ostream& err)
        : Parser(lexer, out, err)
    {
    }
    void run() override;
};

#endif // DUMMY_PARSER_HPP
